'''
See COPYRIGHT.md for copyright information.
'''
from __future__ import annotations
from typing import Any, Sequence, TYPE_CHECKING
import math
from arelle.ModelValue import QName, DateTime
from arelle.ModelObject import ModelObject, ModelAttribute
from arelle.XmlValidateConst import UNKNOWN, VALID, VALID_ID
from arelle.XmlValidate import validate as xmlValidate

if TYPE_CHECKING:
    from arelle.ModelXbrl import ModelXbrl

S_EQUAL: int = 0 # ordinary S-equality from 2.1 spec
S_EQUAL2: int = 1 # XDT definition adds QName comparisons
XPATH_EQ: int = 2 # XPath EQ on all types
VALIDATE_BY_STRING_VALUE: int = 3

NO_IDs_EXCLUDED: int = 0
ALL_IDs_EXCLUDED: int = 1
TOP_IDs_EXCLUDED: int = 2 # only ancestor IDs are excluded

def nodesCorrespond(
    dts1: ModelXbrl,
    elt1: ModelObject | ModelAttribute | None,
    elt2: ModelObject | ModelAttribute | None,
    dts2: ModelXbrl | None = None,
    equalMode: int = XPATH_EQ,
    excludeIDs: int = ALL_IDs_EXCLUDED
) -> bool:
    if elt1 is None:
        return elt2 is None #both can be empty sequences (no element) and true
    elif elt2 is None or not isinstance(elt1, (ModelObject,ModelAttribute)) or not isinstance(elt2, (ModelObject,ModelAttribute)):
        return False
    # can accept either modelElements or modelAttributes
    if isinstance(elt1,ModelAttribute):
        if isinstance(elt2,ModelAttribute):
            return elt1.attrTag == elt2.attrTag and elt1.xValue == elt2.xValue
        else:
            return False
    elif isinstance(elt2,ModelAttribute):
        return False
    # sEqual only accepts modelElements
    return sEqual(dts1, elt1, elt2, equalMode=equalMode, dts2=dts2, excludeIDs=ALL_IDs_EXCLUDED)

# dts1 is modelXbrl for first element
# dts2 is for second element assumed same unless dts2 and ns2 to ns1 mapping table provided
#   (as used in versioning reports and multi instance

def equalityHash(
    elt: ModelObject | Sequence[ModelObject],
    equalMode: int = S_EQUAL,
    excludeIDs: int = NO_IDs_EXCLUDED
) -> int:
    if isinstance(elt, ModelObject):
        try:
            if equalMode == S_EQUAL:
                return elt._hashSEqual
            else:
                return elt._hashXpathEqual
        except AttributeError:
            dts = elt.modelXbrl
            from arelle.ModelXbrl import ModelXbrl
            assert isinstance(dts, ModelXbrl), 'dts is not an instance of ModelXbrl'
            if not hasattr(elt,"xValid"):
                xmlValidate(dts, elt)
            hashableValue = elt.sValue if equalMode == S_EQUAL else elt.xValue
            if isinstance(hashableValue,float) and math.isnan(hashableValue):
                hashableValue = (hashableValue,elt)  # type: ignore[assignment]    # ensure this NaN only compares to itself and no other NaN
            _hash = hash((elt.elementQname,
                          hashableValue,
                          tuple(sorted(attributeDict(dts, elt, set(), equalMode, excludeIDs, distinguishNaNs=True).items(),
                                       key=lambda item: item[0])), # must sort so attrs always hashed in same order
                          tuple(equalityHash(child,equalMode,excludeIDs) for child in childElements(elt))
                          ))
            if equalMode == S_EQUAL:
                elt._hashSEqual = _hash
            else:
                elt._hashXpathEqual = _hash
            return _hash
    elif isinstance(elt, (tuple,list,set)):
        return hash( tuple(equalityHash(i) for i in elt) )
    else:
        return hash(None)

def sEqual(
    dts1: ModelXbrl,
    elt1: ModelObject,
    elt2: ModelObject,
    equalMode: int = S_EQUAL,
    excludeIDs: int = NO_IDs_EXCLUDED,
    dts2: ModelXbrl | None = None,
    ns2ns1Tbl: dict[str, str] | None = None
) -> bool:
    if dts2 is None: dts2 = dts1
    if elt1.localName != elt2.localName:
        return False
    if ns2ns1Tbl and elt2.namespaceURI in ns2ns1Tbl:
        if elt1.namespaceURI != ns2ns1Tbl[elt2.namespaceURI]:
            return False
    elif elt1.namespaceURI != elt2.namespaceURI:
        return False
    if not hasattr(elt1,"xValid"):
        xmlValidate(dts1, elt1)
    if not hasattr(elt2,"xValid"):
        xmlValidate(dts2, elt2)
    children1 = childElements(elt1)
    children2 = childElements(elt2)
    if len(children1) != len(children2):
        return False
    if (not xEqual(elt1, elt2,
                   # must use stringValue for nested contents of mixed content
                   # ... this is now in xValue for mixed content
                   # VALIDATE_BY_STRING_VALUE if len(children1) and elt1.xValid == VALID else
                   equalMode
                   ) or
        attributeDict(dts1, elt1, set(), equalMode, excludeIDs) !=
        attributeDict(dts2, elt2, set(), equalMode, excludeIDs, ns2ns1Tbl)):
        return False
    excludeChildIDs = excludeIDs if excludeIDs != TOP_IDs_EXCLUDED else NO_IDs_EXCLUDED
    for i in range( len(children1) ):
        if not sEqual(dts1, children1[i], children2[i], equalMode, excludeChildIDs, dts2, ns2ns1Tbl):
            return False
    return True

def attributeDict(
    modelXbrl: ModelXbrl,
    elt: ModelObject,
    exclusions: set[str] = set(),
    equalMode: int = S_EQUAL,
    excludeIDs: int = NO_IDs_EXCLUDED,
    ns2ns1Tbl: dict[str, str] | None = None,
    keyByTag: bool = False,
    distinguishNaNs: bool = False
) -> dict[QName, Any]:  # value can be any element value
    if not hasattr(elt,"xValid"):
        xmlValidate(modelXbrl, elt)
    attrs = {}
    # TBD: replace with validated attributes
    for modelAttribute in getattr(elt, 'xAttributes', {}).values():
        attrTag = modelAttribute.attrTag
        ns, sep, localName = attrTag.partition('}')
        attrNsURI = ns[1:] if sep else None
        if ns2ns1Tbl and attrNsURI in ns2ns1Tbl:
            attrNsURI = ns2ns1Tbl[attrNsURI]
        if (attrTag not in exclusions and
            (attrNsURI is None or attrNsURI not in exclusions)):
            if keyByTag:
                qname = attrTag
            elif attrNsURI is not None:
                qname = QName(None, attrNsURI, localName)
            else:
                qname = QName(None, None, attrTag)
            try:
                if excludeIDs and getattr(modelAttribute, "xValid", 0) == VALID_ID:
                    continue
                if modelAttribute.xValid != UNKNOWN:
                    value = modelAttribute.sValue if equalMode <= S_EQUAL2 else modelAttribute.xValue
                else: # unable to validate, no schema definition, use string value of attribute
                    value = modelAttribute.text
                if distinguishNaNs and isinstance(value,float) and math.isnan(value):
                    value = (value,elt)
                attrs[qname] = value
            except KeyError:
                pass  # what should be done if attribute failed to have psvi value
    return attrs

def attributes(
    modelXbrl: ModelXbrl,
    elt: ModelObject,
    exclusions: set[str] = set(),
    ns2ns1Tbl: dict[str, str] | None = None,
    keyByTag: bool = False
) -> tuple[tuple[QName, Any], ...]:
    a = attributeDict(modelXbrl, elt, exclusions, ns2ns1Tbl=ns2ns1Tbl, keyByTag=keyByTag)
    return tuple( (k,a[k]) for k in sorted(a.keys()) )

def childElements(elt: ModelObject) -> list[ModelObject]:
    return [child for child in elt if isinstance(child,ModelObject)]

def xEqual(elt1: ModelObject, elt2: ModelObject, equalMode: int = S_EQUAL) -> bool:
    if not hasattr(elt1,"xValid"):
        xmlValidate(elt1.modelXbrl, elt1)
    if not hasattr(elt2,"xValid"):
        xmlValidate(elt2.modelXbrl, elt2)
    if equalMode == VALIDATE_BY_STRING_VALUE:
        return elt1.stringValue == elt2.stringValue
    elif equalMode == S_EQUAL: # formula WG e-mail 2018-09-06: or (equalMode == S_EQUAL2 and not isinstance(elt1.sValue, QName)):
        return elt1.sValue == elt2.sValue
    else: # includes dimension S-equal2, use xpath-2 equality.
        if isinstance(elt1.xValue, DateTime) \
            and isinstance(elt2.xValue, DateTime) \
                and elt1.xValue.dateOnly != elt2.xValue.dateOnly:
            return False
        return elt1.xValue == elt2.xValue

def vEqual(elt1: ModelObject, elt2: ModelObject) -> bool:
    if not hasattr(elt1,"xValid"):
        xmlValidate(elt1.modelXbrl, elt1)
    if not hasattr(elt2,"xValid"):
        xmlValidate(elt2.modelXbrl, elt2)
    return elt1.sValue == elt2.sValue

def typedValue(
    dts: ModelXbrl | None,
    element: ModelObject,
    attrQname: QName | None = None
) -> Any:  # This can by any type
    try:
        if attrQname: # PSVI attribute value
            modelAttribute = getattr(element, 'xAttributes')[attrQname.clarkNotation]
            if modelAttribute.xValid >= VALID:
                return modelAttribute.xValue
        else: # PSVI element value (of text)
            if getattr(element, 'xValid') >= VALID:
                return element.xValue
    except (AttributeError, KeyError):
        if dts:
            xmlValidate(dts, element, recurse=False, attrQname=attrQname)
            return typedValue(None, element, attrQname=attrQname)
    return None



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

config = ConformanceSuiteConfig(
    args=[
        '--formula', 'run',
        '--httpsRedirectCache',
    ],
    expected_failure_ids=frozenset(f'oim-conformance-2023-04-19/{s}' for s in [
        '600-xml/index-xbrl-xml.xml:V-05',
        '600-xml/index-xbrl-xml.xml:V-06',
    ]),
    file='oim-conformance-2023-04-19/oim-index.xml',
    info_url='https://specifications.xbrl.org/work-product-index-open-information-model-open-information-model.html',
    local_filepath='oim-conformance-2023-04-19.zip',
    membership_url='https://www.xbrl.org/join',
    name=PurePath(__file__).stem,
    plugins=frozenset({'loadFromOIM'}),
)



'''
This module provides database interfaces to postgres SQL

See COPYRIGHT.md for copyright information.
'''
from arelle import XbrlConst

def EFMlinkRoleURIstructure(dts, roleURI):
    relSet = dts.relationshipSet(XbrlConst.parentChild, roleURI)
    dimMems = {} # by dimension qname, set of member qnames
    priItems = set()
    for rootConcept in relSet.rootConcepts:
        EFMlinkRoleDescendants(relSet, rootConcept, dimMems, priItems)
    return dimMems, priItems

def EFMlinkRoleDescendants(relSet, concept, dimMems, priItems):
    if concept is not None:
        if concept.isDimensionItem:
            dimMems[concept.qname] = EFMdimMems(relSet, concept, set())
        else:
            if not concept.isAbstract:
                priItems.add(concept.qname)
            for rel in relSet.fromModelObject(concept):
                EFMlinkRoleDescendants(relSet, rel.toModelObject, dimMems, priItems)

def EFMdimMems(relSet, concept, memQNames):
    for rel in relSet.fromModelObject(concept):
        dimConcept = rel.toModelObject
        if dimConcept is not None and dimConcept.isDomainMember:
            memQNames.add(dimConcept.qname)
            EFMdimMems(relSet, dimConcept, memQNames)
    return memQNames

def tableFacts(dts):
    # identify tables
    disclosureSystem = dts.modelManager.disclosureSystem
    if disclosureSystem.validationType in ("EFM", "HMRC"):
        roleURIcodeFacts = []  # list of (roleURI, code, fact)

        # resolve structural model
        roleTypes = [roleType
                     for roleURI in dts.relationshipSet(XbrlConst.parentChild).linkRoleUris
                     for roleType in dts.roleTypes.get(roleURI,())]
        roleTypes.sort(key=lambda roleType: roleType.definition)
        # find defined non-default axes in pre hierarchy for table
        factsByQname = dts.factsByQname
        for roleType in roleTypes:
            roleURI = roleType.roleURI
            code = roleType.tableCode
            roleURIdims, priItemQNames = EFMlinkRoleURIstructure(dts, roleURI)
            for priItemQName in priItemQNames:
                for fact in factsByQname[priItemQName]:
                    cntx = fact.context
                    # non-explicit dims must be default
                    if (cntx is not None and
                        all(dimQn in dts.qnameDimensionDefaults
                            for dimQn in (roleURIdims.keys() - cntx.qnameDims.keys())) and
                        all(mdlDim.memberQname in roleURIdims[dimQn]
                            for dimQn, mdlDim in cntx.qnameDims.items()
                            if dimQn in roleURIdims)):
                        roleURIcodeFacts.append((roleType, code, fact))

        return roleURIcodeFacts
    return None



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
from arelle.ModelDocument import Type
from arelle.ModelValue import qname
from arelle import XmlUtil, XbrlConst
from arelle.ValidateXbrlCalcs import inferredPrecision, inferredDecimals

def validate(val, modelXbrl, infosetModelXbrl):
    infoset = infosetModelXbrl.modelDocument
    if infoset.type == Type.INSTANCE:
        # compare facts (assumed out of order)
        infosetFacts = defaultdict(list)
        for fact in infosetModelXbrl.facts:
            infosetFacts[fact.qname].append(fact)
        if len(modelXbrl.factsInInstance) != len(infosetModelXbrl.factsInInstance):
            modelXbrl.error("arelle:infosetTest",
                _("Fact counts mismatch, testcase instance %(foundFactCount)s, infoset instance %(expectedFactCount)s"),
                modelObject=(modelXbrl.modelDocument, infosetModelXbrl.modelDocument),
                            foundFactCount=len(modelXbrl.factsInInstance),
                            expectedFactCount=len(infosetModelXbrl.factsInInstance))
        else:
            for i, instFact in enumerate(modelXbrl.facts):
                infosetFact = None
                for fact in infosetFacts[instFact.qname]:
                    if fact.isTuple and fact.isDuplicateOf(instFact, deemP0Equal=True):
                        infosetFact = fact
                        break
                    elif fact.isItem and fact.isVEqualTo(instFact, deemP0Equal=True):
                        infosetFact = fact
                        break
                if infosetFact is None: # takes precision/decimals into account
                    if fact is not None:
                        fact.isVEqualTo(instFact, deemP0Equal=True)
                    modelXbrl.error("arelle:infosetTest",
                        _("Fact %(factNumber)s mismatch %(concept)s"),
                        modelObject=instFact,
                                    factNumber=(i+1),
                                    concept=instFact.qname)
                else:
                    ptvPeriodType = infosetFact.get("{http://www.xbrl.org/2003/ptv}periodType")
                    ptvBalance = infosetFact.get("{http://www.xbrl.org/2003/ptv}balance")
                    ptvDecimals = infosetFact.get("{http://www.xbrl.org/2003/ptv}decimals")
                    ptvPrecision = infosetFact.get("{http://www.xbrl.org/2003/ptv}precision")
                    if ptvPeriodType and ptvPeriodType != instFact.concept.periodType:
                        modelXbrl.error("arelle:infosetTest",
                            _("Fact %(factNumber)s periodType mismatch %(concept)s expected %(expectedPeriodType)s found %(foundPeriodType)s"),
                            modelObject=(instFact, infosetFact),
                                        factNumber=(i+1),
                                        concept=instFact.qname,
                                        expectedPeriodType=ptvPeriodType,
                                        foundPeriodType=instFact.concept.periodType)
                    if ptvBalance and ptvBalance != instFact.concept.balance:
                        modelXbrl.error("arelle:infosetTest",
                            _("Fact %(factNumber)s balance mismatch %(concept)s expected %(expectedBalance)s found %(foundBalance)s"),
                            modelObject=(instFact, infosetFact),
                                        factNumber=(i+1),
                                        concept=instFact.qname,
                                        expectedBalance=ptvBalance,
                                        foundBalance=instFact.concept.balance)
                    if ptvDecimals and ptvDecimals != str(inferredDecimals(fact)):
                        modelXbrl.error("arelle:infosetTest",
                            _("Fact %(factNumber)s inferred decimals mismatch %(concept)s expected %(expectedDecimals)s found %(inferredDecimals)s"),
                            modelObject=(instFact, infosetFact),
                                        factNumber=(i+1),
                                        concept=instFact.qname,
                                        expectedDecimals=ptvDecimals,
                                        inferredDecimals=str(inferredDecimals(fact)))
                    if ptvPrecision and ptvPrecision != str(inferredPrecision(fact)):
                        modelXbrl.error("arelle:infosetTest",
                            _("Fact %(factNumber)s inferred precision mismatch %(concept)s expected %(expectedPrecision)s found %(inferredPrecision)s"),
                            modelObject=(instFact, infosetFact),
                                        factNumber=(i+1),
                                        concept=instFact.qname,
                                        expectedPrecisions=ptvPrecision,
                                        inferredPrecision=str(inferredPrecision(fact)))

    elif infoset.type == Type.ARCSINFOSET:
        # compare arcs
        for arcElt in XmlUtil.children(infoset.xmlRootElement, "http://www.xbrl.org/2003/ptv", "arc"):
            linkType = arcElt.get("linkType")
            arcRole = arcElt.get("arcRole")
            extRole = arcElt.get("extRole")
            fromObj = resolvePath(modelXbrl, arcElt.get("fromPath"))
            if fromObj is None:
                modelXbrl.error("arelle:infosetTest",
                    _("Arc fromPath not found: %(fromPath)s"),
                    modelObject=arcElt, fromPath=arcElt.get("fromPath"))
                continue
            if linkType in ("label", "reference"):
                labelLang = arcElt.get("labelLang")
                resRole = arcElt.get("resRole")
                if linkType == "label":
                    expectedLabel = XmlUtil.text(arcElt)
                    foundLabel = fromObj.label(preferredLabel=resRole,fallbackToQname=False,lang=None,strip=True,linkrole=extRole)
                    if foundLabel != expectedLabel:
                        modelXbrl.error("arelle:infosetTest",
                            _("Label expected='%(expectedLabel)s', found='%(foundLabel)s'"),
                            modelObject=arcElt, expectedLabel=expectedLabel, foundLabel=foundLabel)
                    continue
                elif linkType == "reference":
                    expectedRef = XmlUtil.innerText(arcElt)
                    referenceFound = False
                    for refrel in modelXbrl.relationshipSet(XbrlConst.conceptReference,extRole).fromModelObject(fromObj):
                        ref = refrel.toModelObject
                        if resRole == ref.role:
                            foundRef = XmlUtil.innerText(ref)
                            if foundRef != expectedRef:
                                modelXbrl.error("arelle:infosetTest",
                                    _("Reference inner text expected='%(expectedRef)s, found='%(foundRef)s'"),
                                    modelObject=arcElt, expectedRef=expectedRef, foundRef=foundRef)
                            referenceFound = True
                            break
                    if referenceFound:
                        continue
                modelXbrl.error("arelle:infosetTest",
                    _("%(linkType)s not found containing '%(text)s' linkRole %(linkRole)s"),
                    modelObject=arcElt, linkType=linkType.title(), text=XmlUtil.innerText(arcElt), linkRole=extRole)
            else:
                toObj = resolvePath(modelXbrl, arcElt.get("toPath"))
                if toObj is None:
                    modelXbrl.error("arelle:infosetTest",
                        _("Arc toPath not found: %(toPath)s"),
                        modelObject=arcElt, toPath=arcElt.get("toPath"))
                    continue
                weight = arcElt.get("weight")
                if weight is not None:
                    weight = float(weight)
                order = arcElt.get("order")
                if order is not None:
                    order = float(order)
                preferredLabel = arcElt.get("preferredLabel")
                found = False
                for rel in modelXbrl.relationshipSet(arcRole, extRole).fromModelObject(fromObj):
                    if (rel.toModelObject == toObj and
                        (weight is None or rel.weight == weight) and
                        (order is None or rel.order == order)):
                        found = True
                if not found:
                    modelXbrl.error("arelle:infosetTest",
                        _("Arc not found: from %(fromPath)s, to %(toPath)s, role %(arcRole)s, linkRole $(extRole)s"),
                        modelObject=arcElt, fromPath=arcElt.get("fromPath"), toPath=arcElt.get("toPath"), arcRole=arcRole, linkRole=extRole)
                    continue
        # validate dimensions of each fact
        factElts = XmlUtil.children(modelXbrl.modelDocument.xmlRootElement, None, "*")
        for itemElt in XmlUtil.children(infoset.xmlRootElement, None, "item"):
            try:
                qnElt = XmlUtil.child(itemElt,None,"qnElement")
                factQname = qname(qnElt, XmlUtil.text(qnElt))
                sPointer = int(XmlUtil.child(itemElt,None,"sPointer").text)
                factElt = factElts[sPointer - 1] # 1-based xpath indexing
                if factElt.qname != factQname:
                    modelXbrl.error("arelle:infosetTest",
                        _("Fact %(sPointer)s mismatch Qname, expected %(qnElt)s, observed %(factQname)s"),
                        modelObject=itemElt, sPointer=sPointer, qnElt=factQname, factQname=factElt.qname)
                elif not factElt.isItem or factElt.context is None:
                    modelXbrl.error("arelle:infosetTest",
                        _("Fact %(sPointer)s has no context: %(qnElt)s"),
                        modelObject=(itemElt,factElt), sPointer=sPointer, qnElt=factQname)
                else:
                    context = factElt.context
                    memberElts = XmlUtil.children(itemElt,None,"member")
                    numNonDefaults = 0
                    for memberElt in memberElts:
                        dimElt = XmlUtil.child(memberElt, None, "qnDimension")
                        qnDim = qname(dimElt, XmlUtil.text(dimElt))
                        isDefault = XmlUtil.text(XmlUtil.child(memberElt, None, "bDefaulted")) == "true"
                        if not isDefault:
                            numNonDefaults += 1
                        if not ((qnDim in context.qnameDims and not isDefault) or
                                (qnDim in factElt.modelXbrl.qnameDimensionDefaults and isDefault)):
                            modelXbrl.error("arelle:infosetTest",
                                _("Fact %(sPointer)s (qnElt)s dimension mismatch %(qnDim)s"),
                                modelObject=(itemElt, factElt, context), sPointer=sPointer, qnElt=factQname, qnDim=qnDim)
                    if numNonDefaults != len(context.qnameDims):
                        modelXbrl.error("arelle:infosetTest",
                            _("Fact %(sPointer)s (qnElt)s dimensions count mismatch"),
                            modelObject=(itemElt, factElt, context), sPointer=sPointer, qnElt=factQname)
            except (IndexError, ValueError, AttributeError) as err:
                modelXbrl.error("arelle:infosetTest",
                    _("Invalid entity fact dimensions infoset sPointer: %(test)s, error details: %(error)s"),
                    modelObject=itemElt, test=XmlUtil.innerTextList(itemElt), error=str(err))

def resolvePath(modelXbrl, namespaceId):
    ns, sep, id = (namespaceId or "#").partition("#")
    docs = modelXbrl.namespaceDocs.get(ns)
    if docs: # a list of schema modelDocs with this namespace
        doc = docs[0]
        if id in doc.idObjects:
            return doc.idObjects[id]
    return None

def validateRenderingInfoset(modelXbrl, comparisonFile, sourceDoc):
    from lxml import etree
    try:
        # etree.parse does not successfully parse files within archives
        if modelXbrl.fileSource.isInArchive(comparisonFile):
            file, encoding = modelXbrl.fileSource.file(comparisonFile)
            comparisonDoc = etree.parse(file)
        else:
            comparisonDoc = etree.parse(comparisonFile)
        sourceIter = sourceDoc.iter()
        comparisonIter = comparisonDoc.iter()
        sourceElt = next(sourceIter, None)
        comparisonElt = next(comparisonIter, None)
        # skip over nsmap elements used to create output trees
        while (sourceElt is not None and sourceElt.tag == "nsmap"):
            sourceElt = next(sourceIter, None)
        while (comparisonElt is not None and sourceElt.tag == "nsmap"):
            comparisonElt = next(comparisonIter, None)
        while (sourceElt is not None and comparisonElt is not None):
            while (isinstance(sourceElt, etree._Comment)):
                sourceElt = next(sourceIter, None)
            while (isinstance(comparisonElt, etree._Comment)):
                comparisonElt = next(comparisonIter, None)
            sourceEltTag = sourceElt.tag if sourceElt is not None else '(no more elements)'
            comparisonEltTag = comparisonElt.tag if comparisonElt is not None else '(no more elements)'
            if sourceEltTag != comparisonEltTag:
                modelXbrl.error("arelle:tableModelElementMismatch",
                    _("Table model expecting %(elt1)s found %(elt2)s source line %(elt1line)s comparison line %(elt2line)s"),
                    modelObject=modelXbrl, elt1=sourceEltTag, elt2=comparisonEltTag,
                    elt1line=sourceElt.sourceline, elt2line=comparisonElt.sourceline)
            else:
                text1 = (sourceElt.text or '').strip() or '(none)'
                text2 = (comparisonElt.text or '').strip() or '(none)'
                if text1 != text2:
                    modelXbrl.error("arelle:tableModelTextMismatch",
                        _("Table model comparison element %(elt)s expecting text %(text1)s found %(text2)s source line %(elt1line)s comparison line %(elt2line)s"),
                        modelObject=modelXbrl, elt=sourceElt.tag, text1=text1, text2=text2,
                        elt1line=sourceElt.sourceline, elt2line=comparisonElt.sourceline)
                attrs1 = dict(sourceElt.items())
                attrs2 = dict(comparisonElt.items())
                # remove attributes not to be compared
                for attr in ("{http://www.w3.org/XML/1998/namespace}base",
                             ):
                    if attr in attrs1: del attrs1[attr]
                    if attr in attrs2: del attrs2[attr]
                if attrs1 != attrs2:
                    modelXbrl.error("arelle:tableModelAttributesMismatch",
                        _("Table model comparison element %(elt)s expecting attributes %(attrs1)s found %(attrs2)s source line %(elt1line)s comparison line %(elt2line)s"),
                        modelObject=modelXbrl, elt=sourceElt.tag,
                        attrs1=', '.join('{0}="{1}"'.format(k,v) for k,v in sorted(attrs1.items())),
                        attrs2=', '.join('{0}="{1}"'.format(k,v) for k,v in sorted(attrs2.items())),
                        elt1line=sourceElt.sourceline, elt2line=comparisonElt.sourceline)
            sourceElt = next(sourceIter, None)
            comparisonElt = next(comparisonIter, None)
    except (IOError, etree.LxmlError) as err:
        modelXbrl.error("arelle:tableModelFileError",
            _("Table model comparison file %(xmlfile)s error %(error)s"),
            modelObject=modelXbrl, xmlfile=comparisonFile, error=str(err))



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, N, S, E, W, messagebox
try:
    from tkinter.ttk import Frame, Button
except ImportError:
    from ttk import Frame, Button
import regex as re
from arelle.UiUtil import gridHdr, gridCell, gridCombobox, label, checkbox, radiobutton
from arelle.CntlrWinTooltip import ToolTip
from arelle import ModelDocument, XmlUtil
from arelle.formula import XPathContext, XPathParser
from arelle.ModelDtsObject import ModelConcept
from arelle.ModelInstanceObject import ModelFact
from arelle.ModelRssItem import ModelRssItem
from arelle.ModelFormulaObject import Trace

'''
caller checks accepted, if True, caller retrieves url
'''

reMetaChars = '[]\\^$.|?*+(){}'

newFindOptions = {
    "direction": "down",
    "exprType": "text",
    "all": False,
    "conceptLabel": False,
    "conceptName": False,
    "conceptSubs": False,
    "conceptPer": False,
    "conceptBal": False,
    "factLabel": False,
    "factName": False,
    "factValue": False,
    "factCntx": False,
    "factUnit": False,
    "messagesLog": False,
    "priorExpressions": [],
    "geometry": None
}

def find(mainWin):
    dialog = DialogFind(mainWin, mainWin.config.setdefault("findOptions", newFindOptions))


class DialogFind(Toplevel):
    def __init__(self, mainWin, options):
        parent = mainWin.parent
        super(DialogFind, self).__init__(parent)
        self.parent = parent
        self.modelManager = mainWin.modelManager
        self.modelXbrl = None   # set when Find pressed, this blocks next prematurely
        if options is None: options = newFindOptions
        self.options = options
        parentGeometry = re.match("(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogW = int(parentGeometry.group(1))
        dialogH = int(parentGeometry.group(2))
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.accepted = False

        self.transient(self.parent)
        self.title(_("Find"))

        self.objsList = [] # next may be tried before anything is found

        frame = Frame(self)

        # load grid
        findLabel = gridHdr(frame, 1, 0, "Find:", anchor="w")
        findLabel.grid(padx=8)
        self.cbExpr = gridCombobox(frame, 1, 1, values=options["priorExpressions"])
        self.cbExpr.grid(columnspan=3, padx=8)
        ToolTip(self.cbExpr, text=_("Enter expression to find, or select from combo box drop down history list."), wraplength=240)

        y = 2

        # checkbox entries
        label(frame, 1, y, "Direction:")
        label(frame, 1, y + 3, "Match:")
        scopeLabel = label(frame, 2, y, "Scope:")
        ToolTip(scopeLabel, text=_("Scope for an XBRL document (instance or DTS).  "
                                   "For an RSS Feed, all properties are matched.  "), wraplength=240)
        rbUp = radiobutton(frame, 1, y+1, "Up", "up", "direction")
        ToolTip(rbUp, text=_("Find/Next up (on screen) from last to first match."), wraplength=240)
        rbDn = radiobutton(frame, 1, y+2, "Down", "down", "direction", rbUp.valueVar)
        ToolTip(rbDn, text=_("Find/Next down (on screen) from first to last match."), wraplength=240)
        rbText = radiobutton(frame, 1, y+4, "Text (ignore case)", "text", "exprType")
        ToolTip(rbText, text=_("Expression is a set of characters to match, ignoring case.  "
                               "The match may occur anywhere within the scope. "), wraplength=360)
        rbRegex = radiobutton(frame, 1, y+5, "Regular expression", "regex", "exprType", rbText.valueVar)
        ToolTip(rbRegex, text=_('A regular expression to match, anywhere in the scope, ignoring case.  '
                                'For example, "cash" would match cash anywhere in a string (like cash on hand), '
                                'whereas "^cash$" would match a full string to only contain cash. '
                                'Use regular expression metacharacters, e.g., "." for any single character, '
                                '".*" for any number of wild characters, .{3} for exactly 3 wild characters. '), wraplength=360)
        rbXPath = radiobutton(frame, 1, y+6, "XPath 2 expression", "xpath", "exprType", rbText.valueVar)
        ToolTip(rbXPath, text=_('An XPath 2 expression, where the context element, ".", is a candidate concept QName, if any concept scope is checked, '
                                'and a candidate fact item, if any fact scope is checked.  The XPath 2 functions do not need an "fn:" prefix (but it is defined).  '
                                'The XBRL Functions Registry functions do require an "xfi:" prefix.  Constructors require an "xs:" prefix.  '
                                'The expression is considered "matched" for the candidate concept QNames or fact items where the effective boolean value of the expression is "true()".  '), wraplength=360)
        self.optionControls = (
           rbUp,
           rbDn,
           rbText,
           rbRegex,
           rbXPath,
           #checkbox(frame, 2, y + 1, "All", "all"),
           checkbox(frame, 2, y + 1, "Concept label", "conceptLabel"),
           checkbox(frame, 2, y + 2, "   name", "conceptName"),
           checkbox(frame, 2, y + 3, "   type", "conceptType"),
           checkbox(frame, 2, y + 4, "   subs group", "conceptSubs"),
           checkbox(frame, 2, y + 5, "   period type", "conceptPer"),
           checkbox(frame, 2, y + 6, "   balance", "conceptBal"),
           checkbox(frame, 3, y + 1, "Fact label", "factLabel"),
           checkbox(frame, 3, y + 2, "   name", "factName"),
           checkbox(frame, 3, y + 3, "   value", "factValue"),
           checkbox(frame, 3, y + 4, "   context", "factCntx"),
           checkbox(frame, 3, y + 5, "   unit", "factUnit"),
           checkbox(frame, 3, y + 6, "Messages", "messagesLog"),

           # Note: if adding to this list keep Finder.FindOptions in sync

           )
        y += 7
        resultLabel = gridHdr(frame, 1, y, "Result:", anchor="w")
        resultLabel.grid(padx=8)
        self.resultText = gridCell(frame, 1, y + 1)
        self.resultText.grid(columnspan=3, padx=8)
        self.resultText.config(state="readonly")
        y += 2

        mainWin.showStatus(None)

        buttonFrame = Frame(frame)
        buttonFrame.grid(columnspan=4, sticky=E, padx=8)
        findButton = Button(buttonFrame, text=_("Find"), width=12, command=self.find)
        ToolTip(findButton, text=_('Compile (if regular expression or XPath 2), and find first match (if down direction) or last match (if up direction).  '), wraplength=240)
        nextButton = Button(buttonFrame, text=_("Next"), width=12, command=self.next)
        ToolTip(nextButton, text=_('Advance to the next matched object (in selected direction).  '), wraplength=240)
        closeButton = Button(buttonFrame, text=_("Close"), width=12, command=self.close)
        ToolTip(closeButton, text=_('Close the find dialog.  '), wraplength=240)
        findButton.grid(row=1, column=1, pady=3)
        nextButton.grid(row=1, column=2, pady=3)
        closeButton.grid(row=1, column=3, padx=3)

        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(1, weight=1)
        frame.columnconfigure(2, weight=1)
        frame.columnconfigure(3, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        if self.options["geometry"]:
            self.geometry(self.options["geometry"])
        else:
            self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        #self.bind("<Return>", self.ok)
        #self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)

        # make this dialog non-modal
        self.focus_set()
        #self.grab_set()
        #self.wait_window(self)

    def setOptions(self):
        # set formula options
        for optionControl in self.optionControls:
            self.options[optionControl.attr] = optionControl.value

    def find(self, event=None):
        self.setOptions()
        self.accepted = True
        # self.close()

        docType = self.modelManager.modelXbrl.modelDocument.type if self.modelManager.modelXbrl else None
        if self.options["messagesLog"]:
            if docType == ModelDocument.Type.RSSFEED and self.options["exprType"] == "xpath":
                messagebox.showerror(_("Find cannot be completed"),
                         _("XPath matching is not available for searching messages, please choose text or regular expression.  "), parent=self)
                return
        else:
            if not self.modelManager.modelXbrl or not docType in (
                 ModelDocument.Type.SCHEMA, ModelDocument.Type.LINKBASE, ModelDocument.Type.INSTANCE, ModelDocument.Type.INLINEXBRL,
                 ModelDocument.Type.RSSFEED):
                messagebox.showerror(_("Find cannot be completed"),
                         _("Find requires an opened DTS or RSS Feed"), parent=self.parent)
                return

            if docType == ModelDocument.Type.RSSFEED and self.options["exprType"] == "xpath":
                messagebox.showerror(_("Find cannot be completed"),
                         _("XPath matching is not available for an RSS Feed, please choose text or regular expression.  "), parent=self)
                return

        self.modelXbrl = self.modelManager.modelXbrl
        expr = self.cbExpr.value
        # update find expressions history
        if expr in self.options["priorExpressions"]:
            self.options["priorExpressions"].remove(expr)
        elif len(self.options["priorExpressions"]) > 10:
            self.options["priorExpressions"] = self.options["priorExpressions"][0:10]
        self.options["priorExpressions"].insert(0, expr)
        self.cbExpr.config(values=self.options["priorExpressions"])
        self.saveConfig()

        import threading
        thread = threading.Thread(target=lambda
                                  expr=self.cbExpr.value,
                                  logViewLines=self.modelManager.cntlr.logView.lines() if self.options["messagesLog"] else None
                                  : self.backgroundFind(expr, logViewLines))
        thread.daemon = True
        thread.start()


    def backgroundFind(self, expr, logViewLines):
        exprType = self.options["exprType"]
        inConceptLabel = self.options["conceptLabel"]
        inConceptName = self.options["conceptName"]
        inConceptType = self.options["conceptType"]
        inConceptSubs = self.options["conceptSubs"]
        inConceptPer = self.options["conceptPer"]
        inConceptBal = self.options["conceptBal"]
        inFactLabel = self.options["factLabel"]
        inFactName = self.options["factName"]
        inFactValue = self.options["factValue"]
        inFactCntx = self.options["factCntx"]
        inFactUnit = self.options["factUnit"]
        inMessagesLog = self.options["messagesLog"]
        nextIsDown = self.options["direction"] == "down"

        objsFound = set()

        self.result = "Found "


        try:
            if exprType == "text":
                # escape regex metacharacters
                pattern = re.compile(''.join(
                         [(('\\' + c) if c in reMetaChars else c) for c in expr]),
                         re.IGNORECASE)
                isRE = True
                isXP = False
            elif exprType == "regex":
                pattern = re.compile(expr, re.IGNORECASE)
                isRE = True
                isXP = False
            elif exprType == "xpath":
                isRE = False
                isXP = True
                self.resultText.setValue(_("Compiling xpath expression..."))
                XPathParser.initializeParser(self.modelManager)
                self.modelManager.showStatus(_("Compiling xpath expression..."))
                xpProg= XPathParser.parse(self,
                                          expr,
                                          XPathParser.staticExpressionFunctionContext(),
                                          "find expression",
                                          Trace.CALL)
                xpCtx = XPathContext.create(self.modelXbrl, sourceElement=None)

            else:
                return  # nothing to do

            if inMessagesLog:
                for lineNumber, line in enumerate(logViewLines):
                    if pattern.search(line):
                        objsFound.add(lineNumber)
            elif self.modelXbrl.modelDocument.type == ModelDocument.Type.RSSFEED:
                for rssItem in self.modelXbrl.modelDocument.items:
                    if any(pattern.search(str(value)) for name, value in rssItem.propertyView):
                        objsFound.add(rssItem)
            else: # DTS search
                if inConceptLabel or inConceptName or inConceptType or inConceptSubs or inConceptPer or inConceptBal:
                    self.modelManager.cntlr.uiThreadQueue.put((self.resultText.setValue, [_("Matching concepts...")]))
                    self.modelManager.showStatus(_("Matching concepts..."))
                    for conceptName, concepts in self.modelXbrl.nameConcepts.items():
                        for concept in concepts:
                            if ((isXP and xpCtx.evaluateBooleanValue(xpProg, contextItem=concept.qname)) or
                                (isRE and
                                 (inConceptLabel and pattern.search(concept.label())) or
                                 (inConceptName and pattern.search(conceptName)) or
                                 (inConceptType and pattern.search(str(concept.typeQname))) or
                                 (inConceptSubs and pattern.search(str(concept.substitutionGroupQname))) or
                                 (inConceptPer and concept.periodType and pattern.search(concept.periodType)) or
                                 (inConceptBal and concept.balance and pattern.search(concept.balance))
                                 )
                                ):
                                objsFound.add(concept)
                if inFactLabel or inFactName or inFactValue or inFactCntx or inFactUnit:
                    self.modelManager.cntlr.uiThreadQueue.put((self.resultText.setValue, [_("Matching facts...")]))
                    self.modelManager.showStatus(_("Matching facts..."))
                    for fact in self.modelXbrl.facts:
                        if ((isXP and xpCtx.evaluateBooleanValue(xpProg, contextItem=fact)) or
                            (isRE and
                             (inFactName and pattern.search(fact.concept.name) or
                             (inFactLabel and pattern.search(fact.concept.label())) or
                             (inFactValue and pattern.search(fact.value)) or
                             (inFactCntx and pattern.search(XmlUtil.innerText(fact.context.element))) or
                             (inFactUnit and pattern.search(XmlUtil.innerText(fact.unit.element))))
                             )
                            ):
                            objsFound.add(fact)
        except (XPathContext.XPathException, TypeError, ValueError, OverflowError, IndexError, KeyError, re.error) as err:
            err = _("Find expression error: {0} \n{1}").format(
                str(err),
                getattr(err, "sourceErrorIndication",getattr(err, "pattern","")))
            self.modelManager.addToLog(err)
            self.modelManager.cntlr.uiThreadQueue.put((self.resultText.setValue, [err]))
            self.modelManager.showStatus(_("Completed with errors"), 5000)
            self.result = err + "\n"

        numConcepts = 0
        numFacts = 0
        numRssItems = 0
        numMessages = 0
        self.objsList = []
        for obj in objsFound:
            if inMessagesLog:
                numMessages += 1
                self.objsList.append( ('m', "{0:06}".format(obj), obj) )
            elif isinstance(obj,ModelConcept):
                numConcepts += 1
                self.objsList.append( ('c', obj.localName, obj.objectId()) )
            elif isinstance(obj,ModelFact):
                numFacts += 1
                self.objsList.append( ('f', obj.__hash__(), obj.objectId()) )
            elif isinstance(obj,ModelRssItem):
                numRssItems += 1
                self.objsList.append( ('r', obj.__hash__(), obj.objectId()) )
        self.objsList.sort()
        if numConcepts:
            self.result += "{0} concepts".format(numConcepts)
            if numFacts: self.result += ", "
        if numFacts:
            self.result += "{0} facts".format(numFacts)
        if numRssItems:
            self.result += "{0} RSS items".format(numRssItems)
        if numMessages:
            self.result += "{0} Messages".format(numMessages)
        if numConcepts + numFacts + numRssItems + numMessages == 0:
            self.result += "no matches"
            self.foundIndex = -1
            self.modelManager.cntlr.uiThreadQueue.put((self.resultText.setValue, [self.result]))
        else:
            self.foundIndex = 0 if nextIsDown else (len(self.objsList) - 1)
            self.modelManager.cntlr.uiThreadQueue.put((self.next, []))
        self.modelManager.showStatus(_("Ready..."), 2000)

    def next(self):
        self.setOptions() # refresh options
        nextIsDown = self.options["direction"] == "down"
        # check that asme instance applies
        if not self.options["messagesLog"]:
            if self.modelXbrl is None:
                return
            if self.modelManager.modelXbrl != self.modelXbrl:
                messagebox.showerror(_("Next cannot be completed"),
                                _("A different DTS is active, than find was initiated with.  Please press 'find' to re-search with the current DTS"), parent=self)
                return
        lenObjsList = len(self.objsList)
        if lenObjsList == 0:
            messagebox.showwarning(_("Next cannot be completed"),
                            _("No matches were found.  Please try a different search."), parent=self)
            return

        if self.foundIndex < 0 and nextIsDown:
            self.foundIndex += 1
        elif self.foundIndex >= lenObjsList and not nextIsDown:
            self.foundIndex -= 1
        if 0 <= self.foundIndex < lenObjsList:
            objectFound = self.objsList[self.foundIndex][2]
            if self.options["messagesLog"]:
                self.modelManager.cntlr.logView.selectLine(objectFound)
            else:
                self.modelManager.modelXbrl.viewModelObject(objectFound)
            self.resultText.setValue("{0}, selection {1} of {2}".format(self.result, self.foundIndex + 1, len(self.objsList) ) )
            self.foundIndex += 1 if nextIsDown else -1
        elif nextIsDown:
            self.resultText.setValue("{0}, selection at end".format(self.result) )
        else:
            self.resultText.setValue("{0}, selection at start".format(self.result) )


    def close(self, event=None):
        self.options["geometry"] = self.geometry()
        self.saveConfig()
        self.parent.focus_set()
        self.destroy()

    def saveConfig(self):
        self.modelManager.cntlr.config["findOptions"] = self.options
        self.modelManager.cntlr.saveConfig()



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

TIMING = {f'esef_conformance_suite_2022/tests/inline_xbrl/{k}': v for k, v in {
    'RTS_Annex_III_Par_3_G3-1-3/index.xml': 4.109,
    'RTS_Art_6_a/index.xml': 6.795,
    'RTS_Annex_IV_Par_6/index.xml': 7.087,
    'G2-4-1_3/index.xml': 7.237,
    'RTS_Annex_II_Par_1_RTS_Annex_IV_par_7/index.xml': 7.343,
    'G3-5-1/index.xml': 7.421,
    'RTS_Annex_IV_Par_4_2/index.xml': 7.7,
    'G3-4-2_3/index.xml': 7.762,
    'G2-4-2_1/index.xml': 7.802,
    'G2-5-3/index.xml': 7.83,
    'G3-4-7/index.xml': 7.832,
    'G2-4-1_2/index.xml': 7.855,
    'G3-4-2_1/index.xml': 7.861,
    'G2-7-1_2/index.xml': 7.875,
    'RTS_Annex_IV_Par_14_G2-5-1/index.xml': 7.884,
    'G2-4-2_2/index.xml': 7.922,
    'G2-7-1_1/index.xml': 7.963,
    'G2-6-2/index.xml': 7.99,
    'G2-5-4_2/index.xml': 8.023,
    'RTS_Annex_IV_Par_4_1/index.xml': 8.063,
    'G3-2-2/index.xml': 8.147,
    'G2-5-4_1/index.xml': 8.293,
    'G3-4-2_2/index.xml': 8.305,
    'G2-5-4_3/index.xml': 8.529,
    'RTS_Annex_IV_Par_1_G2-1-4/index.xml': 8.536,
    'G3-2-3/index.xml': 8.569,
    'G3-4-4/index.xml': 8.718,
    'G3-4-5_1/index.xml': 8.763,
    'RTS_Annex_IV_Par_8_G3-4-5/index.xml': 10.151,
    'RTS_Annex_IV_Par_5/index.xml': 10.481,
    'RTS_Annex_IV_Par_2_G2-1-1/index.xml': 11.47,
    'G2-6-1_1/index.xml': 11.625,
    'G2-6-1_2/index.xml': 11.886,
    'RTS_Annex_IV_Par_11_G3-2-2/index.xml': 11.911,
    'G3-4-6/index.xml': 11.955,
    'G3-1-1_2/index.xml': 11.998,
    'RTS_Art_3/index.xml': 12.226,
    'G2-5-1_3/index.xml': 12.396,
    'G3-4-3_2/index.xml': 12.398,
    'G3-4-3_1/index.xml': 12.521,
    'G2-2-2/index.xml': 12.644,
    'G2-2-1/index.xml': 12.648,
    'RTS_Annex_II_Par_1/index.xml': 12.654,
    'G2-1-3_1/index.xml': 12.657,
    'G3-4-2_4/index.xml': 12.821,
    'G2-1-3_2/index.xml': 12.828,
    'G3-4-5_2/index.xml': 12.907,
    'RTS_Annex_III_Par_1/index.xml': 13.035,
    'G2-3-1_1/index.xml': 13.898,
    'G2-4-1_1/index.xml': 15.855,
    'G2-5-1_2/index.xml': 15.948,
    'G2-2-6_2/index.xml': 15.961,
    'G2-5-1_1/index.xml': 16.026,
    'RTS_Annex_IV_Par_4_G1-1-1_G3-4-5/index.xml': 16.173,
    'G2-5-2/index.xml': 16.174,
    'G2-1-2/index.xml': 20.047,
    'G2-2-6_1/index.xml': 20.552,
    'RTS_Annex_IV_Par_9_Par_10_G1-4-1_G1-4-2_G3-3-1_G3-3-2/index.xml': 22.756,
    'RTS_Annex_IV_Par_12_G2-2-4/index.xml': 23.328,
    'G2-3-1_2/index.xml': 23.504,
    'G3-1-1_1/index.xml': 27.177,
    'G3-1-2/index.xml': 27.707,
    'G2-2-3/index.xml': 28.185,
    'G3-1-5/index.xml': 31.773,
}.items()}

config = ConformanceSuiteConfig(
    approximate_relative_timing=TIMING,
    args=[
        '--disclosureSystem', 'esef',
        '--formula', 'run',
    ],
    expected_failure_ids=frozenset(f'esef_conformance_suite_2022/tests/{s}' for s in [
        # The following test cases fail because of the `tech_duplicated_facts1` formula which fires
        # incorrectly because it does not take into account the language attribute on the fact.
        # A fact can not be a duplicate fact if the language attributes are different.
        'inline_xbrl/RTS_Annex_IV_Par_12_G2-2-4/index.xml:TC5_valid'
    ]),
    file='esef_conformance_suite_2022/index_inline_xbrl.xml',
    info_url='https://www.esma.europa.eu/document/conformance-suite-2022',
    local_filepath='esef_conformance_suite_2022.zip',
    name=PurePath(__file__).stem,
    plugins=frozenset({'validate/ESEF_2022'}),
    public_download_url='https://www.esma.europa.eu/sites/default/files/library/esef_conformance_suite_2022.zip',
    shards=5,
)



from arelle.BetaFeatures import newObjectModelEnabled

# The if statement is negated here so that from mypy's perspective the types
# inherit from the lxml object model until the new models are ready to be used.
if not newObjectModelEnabled():
    from lxml.etree import CommentBase as CommentBase
    from lxml.etree import ElementBase as ElementBase
    from lxml.etree import PIBase as PIBase
else:
    from .CommentBase import CommentBase
    from .ElementBase import ElementBase  # type: ignore[assignment]
    from .PIBase import PIBase



from arelle.formula.FactAspectsCache import FactAspectsCache


class TestFactAspectsCache:
    def test_match(self):
        cache = FactAspectsCache(10)
        cache.cacheMatch("fact1", "fact2", "aspect")

        fact1_evaluations = cache.evaluations("fact1", "fact2")
        fact2_evaluations = cache.evaluations("fact2", "fact1")

        assert all(evaluations == {"aspect": True} for evaluations in (fact1_evaluations, fact2_evaluations))

    def test_cache_none_values(self):
        cache = FactAspectsCache(10)

        cache.cacheNotMatch(None, "fact2", "aspect")
        cache.cacheNotMatch("fact1", None, "aspect")
        cache.cacheNotMatch("fact1", "fact2", None)

        fact1_2_evaluations = cache.evaluations("fact1", "fact2")
        fact1_none_evaluations = cache.evaluations("fact1", None)
        fact2_none_evaluations = cache.evaluations(None, "fact2")

        assert fact1_2_evaluations == {
            None: False,
        }

        assert fact1_none_evaluations == {
            "aspect": False,
        }

        assert fact2_none_evaluations == {
            "aspect": False,
        }

    def test_non_match(self):
        cache = FactAspectsCache(10)
        cache.cacheNotMatch("fact1", "fact2", "aspect")

        fact1_evaluations = cache.evaluations("fact1", "fact2")
        fact2_evaluations = cache.evaluations("fact2", "fact1")

        assert all(evaluations == {"aspect": False} for evaluations in (fact1_evaluations, fact2_evaluations))

    def test_mixed_evaluations(self):
        cache = FactAspectsCache(10)
        cache.cacheMatch("fact1", "fact2", "aspect1")
        cache.cacheNotMatch("fact1", "fact2", "aspect2")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            "aspect1": True,
            "aspect2": False,
        }

    def test_empty_cache(self):
        cache = FactAspectsCache(10)

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations is None

    def test_additional_facts(self):
        cache = FactAspectsCache(10)

        cache.cacheMatch("fact1", "fact2", "aspect1")
        cache.cacheNotMatch("fact1", "fact2", "aspect2")
        cache.cacheMatch("fact1", "fact2", "aspect3")

        cache.cacheNotMatch("fact1", "fact3", "aspect1")
        cache.cacheMatch("fact1", "fact3", "aspect2")

        cache.cacheNotMatch("fact2", "fact3", "aspect1")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            "aspect1": True,
            "aspect2": False,
            "aspect3": True,
        }

    def test_max_size_reached(self):
        cache = FactAspectsCache(2)
        cache.cacheNotMatch("fact1", "fact2", "aspect1")
        cache.cacheMatch("fact1", "fact2", "aspect2")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            "aspect1": False,
            "aspect2": True,
        }
        assert cache.prioritizedAspects == {"aspect1"}

        cache.cacheNotMatch("fact1", "fact2", "aspect3")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            "aspect1": False,
            "aspect2": True,
        }
        assert cache.prioritizedAspects == {"aspect1", "aspect3"}

    def test_negative_max_size(self):
        cache = FactAspectsCache(-1)
        for i in range(100):
            cache.cacheNotMatch("fact1", "fact2", f"aspect{i}")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            f"aspect{i}": False
            for i in range(100)
        }
        assert cache.prioritizedAspects == {
            f"aspect{i}"
            for i in range(100)
        }

    def test_zero_max_size(self):
        cache = FactAspectsCache(0)
        cache.cacheNotMatch("fact1", "fact2", "aspect")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations is None
        assert cache.prioritizedAspects == {"aspect"}

    def test_clear(self):
        cache = FactAspectsCache(10)
        cache.cacheMatch("fact1", "fact2", "aspect")

        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations == {
            "aspect": True,
        }

        cache.clear()
        evaluations = cache.evaluations("fact1", "fact2")

        assert evaluations is None

    def test_prioritized_aspects(self):
        cache = FactAspectsCache(10)
        cache.cacheNotMatch("fact1", "fact2", "aspect1")
        cache.cacheNotMatch("fact1", "fact2", "aspect2")
        cache.cacheMatch("fact1", "fact2", "aspect3")
        cache.cacheMatch("fact1", "fact2", "aspect4")
        cache.cacheNotMatch("fact3", "fact4", "aspect5")

        assert cache.prioritizedAspects == {"aspect1", "aspect2", "aspect5"}

    def test_prioritized_aspects_clear(self):
        cache = FactAspectsCache(10)
        cache.cacheNotMatch("fact1", "fact2", "aspect1")

        assert cache.prioritizedAspects == {"aspect1"}

        cache.clear()

        assert cache.prioritizedAspects == set()

        cache.cacheNotMatch("fact1", "fact2", "aspect1")

        assert cache.prioritizedAspects == {"aspect1"}



'''
See COPYRIGHT.md for copyright information.
'''
import os
from collections import defaultdict
from arelle import XbrlConst
from arelle.ModelDtsObject import ModelConcept

# check if concept is behaving as a total based on role, deed, or circumstances
def presumptionOfTotal(val, rel, siblingRels, iSibling, isStatementSheet, nestedInTotal, checkLabelRoleOnly):
    """
    A numeric concept target of a parent-child relationship is presumed total if:

    (i) its preferredLabel role is a total role (pre XbrlConst static function of
    current such total roles) or

    (ii) if not in a nested total (abstract child relationship to a known total's
    contributing siblings):

    the parent is not SupplementalCashFlowInformationAbstract and the preceding
    sibling relationship is monetary and it's on a statement sheet and it's the
    last of more than one monetary item

    (a) Last monetary parented by an abstract or non-monetary and not in a nested
    (breakdown) total, or
    (b) effective label (en-US of preferred role) has "Total" in its wording.
    (c) (commented out for now due to false positives: Concept name has "Total"
    in its name)
    (d) last monetary (may be sub level) whose immediate sibling is a calc LB child
    """
    concept = rel.toModelObject
    if isinstance(concept, ModelConcept) and concept.isNumeric:
        preferredLabel = rel.preferredLabel
        if XbrlConst.isTotalRole(preferredLabel):
            return _("preferredLabel {0}").format(os.path.basename(preferredLabel))
        if concept.isMonetary and not checkLabelRoleOnly:
            effectiveLabel = concept.label(lang="en-US", fallbackToQname=False, preferredLabel=preferredLabel)
            ''' word total in label/name does not seem to be a good indicator,
                e.g., Google Total in label for ShareBasedCompensationArrangementByShareBasedPaymentAwardGrantDateFairValueOfOptionsVested followed by
                label with Aggregate but name has Total
                ... so only perform this test on last monetary in a Note
            if 'Total' in effectiveLabel: # also check for Net ???
                return _("word 'Total' in effective label {0}").format(effectiveLabel)
            if 'Total' in concept.name: # also check for Net ???
                return _("word 'Total' in concept name {0}").format(concept.name)
            '''
            parent = rel.fromModelObject
            if (len(siblingRels) > 1 and
                iSibling == len(siblingRels) - 1 and
                parent is not None and
                parent.name not in {
                    "SupplementalCashFlowInformationAbstract"
                }):
                preceedingSibling = siblingRels[iSibling - 1].toModelObject
                if preceedingSibling is not None and preceedingSibling.isMonetary:
                    # last fact, may be total
                    if isStatementSheet:
                        # check if facts add up??
                        if (parent.isAbstract or not parent.isMonetary) and not nestedInTotal:
                            return _("last monetary item in statement sheet monetary line items parented by nonMonetary concept")
                        elif effectiveLabel and 'Total' in effectiveLabel:
                            return _("last monetary item in statement sheet monetary line items with word 'Total' in effective label {0}").format(effectiveLabel)
                        elif 'Total' in concept.name:
                            return _("last monetary item in statement sheet monetary line items with word 'Total' in concept name {0}").format(concept.name)
                        elif val.summationItemRelsSetAllELRs.isRelated(concept, "child", preceedingSibling):
                            return _("last monetary item in statement sheet monetary line items is calc sum of previous line item")
                    ''' for now unreliable to use total words for notes
                    else:
                        if 'Total' in effectiveLabel: # also check for Net ???
                            return _("last monetary item in note with word 'Total' in effective label {0}").format(effectiveLabel)
                        if 'Total' in concept.name: # also check for Net ???
                            return _("last monetary item in note with word 'Total' in concept name {0}").format(concept.name)
                    '''
    return None

# 6.15.02, 6.15.03
def checkCalcsTreeWalk(val, parentChildRels, concept, isStatementSheet, inNestedTotal, conceptsUsed, visited):
    """
    -  EFM-strict validation 6.15.2/3: finding presumed totals in presentation and inspecting for
       equivalents in calculation (noted as error-semantic, in efm-strict mode).

    -  Best practice approach: inspecting for calcuations in the UGT calculations that would hint
       that like filing constructs should have presentation (noted as warning-semantic in best practices plug-in, when loaded and enabled)

    EFM-strict missing-calcs

    a. Presumption of total

    The presentation linkbase is tree-walked to find items presumed to be totals and their contributing
    items.  (see description of presumptionOfTotal, above)

    b. Finding calculation link roles with least mis-fit to presumed total and its contributing items
    (presumptionOfTotal in ValidateFiling.py).

    For each presumed total (checkForCalculations in ValidateFiling.py):

    b.1 Contributing items are found for the presumed total as follows:

    From the presumed total, walking back through its preceding sibilings (with caution to avoid
    looping on allowed direct cycles), a preceding sibling is a contributing item if it has facts,
    same period type, and numeric.  If a preceding sibling is abstract, the abstract's children are
    likewise recursively checked (as they often represent a breakdown, and such children of an
    abstract sibling to the total are also contributing items (except for such children preceding
    a total at the child level).

    If a preceding sibling is presumed total (on same level), it is a running subtotal (in subsequent
    same-level total) unless it's independent in the calc LB (separate totaled stuff preceding these
    siblings) or related to grandparent sum.

    b.2 Finding the facts of these total/contributing item sets

    Sets of total and compatible contributing facts that match the sets of total concept and
    contributing concept must next be found, because each of these different sets (of total
    and compatible contributing facts) may fit different calculation link roles (according to
    which compatible contributing facts are present for each total).  This is particularly
    important when totals and contributing items exist both on face statements and notes, but
    the contributing compatible fact population is different).

    For each fact of the total concept, that has a specified end/instant datetime and unit, if
    (i) it's not on a statement or
    (ii) required context is absent or
    (iii) the fact's end/instant is within the required context's duration, the contributing
    item facts are those unit and context equivalent to such total fact.

    b.3 Finding least-mis-matched calculation link role

    Each link role in calculation produces a different set of summation-item arc-sets, and
    each set of presumed-total facts and compatible contributing item facts is separately
    considered to find the least-mismatched calculation summation-item arc-set.

    The link roles are not intermixed or aggregated, each link role produces independent
    summation-item arc-sets (XBRL 2.1 section 5.2.5.2).

    For each total fact and compatible contributing item facts, the calculation link roles
    are examined one-by-one for that link-role where the total has children missing the
    least of the compatible contributing item fact children, and reported either as 6.15.02
    (for statement sheet presentation link roles) or 6.15.03 (for non-statement link roles).
    The determination of statement sheet is according to the presentation tree walk.  The
    search for least-misfit calculation link role does not care or consider the value of the
    calculation link role, just the summation-item arc-set from the presumed-total concept.
    """
    if concept not in visited:
        visited.add(concept)
        siblingRels = parentChildRels.fromModelObject(concept)
        foundTotalAtThisLevel = False
        for iSibling, rel in enumerate(siblingRels):
            reasonPresumedTotal = presumptionOfTotal(val, rel, siblingRels, iSibling, isStatementSheet, False, inNestedTotal)
            if reasonPresumedTotal:
                foundTotalAtThisLevel = True
                checkForCalculations(val, parentChildRels, siblingRels, iSibling, rel.toModelObject, rel, reasonPresumedTotal, isStatementSheet, conceptsUsed, False, set())
        if foundTotalAtThisLevel: # try nested tree walk to look for lower totals
            inNestedTotal = True
        for rel in siblingRels:
            checkCalcsTreeWalk(val, parentChildRels, rel.toModelObject, isStatementSheet, inNestedTotal, conceptsUsed, visited)
        visited.remove(concept)

def checkForCalculations(val, parentChildRels, siblingRels, iSibling, totalConcept, totalRel, reasonPresumedTotal, isStatementSheet, conceptsUsed, nestedItems, contributingItems):
    # compatible preceding sibling facts must have calc relationship to toal
    for iContributingRel in range(iSibling - 1, -1, -1):
        contributingRel = siblingRels[iContributingRel]
        siblingConcept = contributingRel.toModelObject
        if siblingConcept is not None:
            if siblingConcept is totalConcept: # direct cycle loop likely, possibly among children of abstract sibling
                break
            if val.summationItemRelsSetAllELRs.isRelated(totalConcept, 'ancestral-sibling', siblingConcept):
                break # sibling independently contributes as sibling of totalConcept to grandfather total
            if any(val.summationItemRelsSetAllELRs.isRelated(contributingItem, 'child', siblingConcept)
                   for contributingItem in contributingItems):
                break # this subtotal is a breakdown of something already being considered
            isContributingTotal = presumptionOfTotal(val, contributingRel, siblingRels, iContributingRel, isStatementSheet, True, False)
            # contributing total may actually be separate non-running subtotal, if so don't include it here
            if isContributingTotal:
                if (val.summationItemRelsSetAllELRs.fromModelObject(siblingConcept) and not
                    val.summationItemRelsSetAllELRs.toModelObject(siblingConcept)):
                    break # sibling independently contributes as sibling of totalConcept as a root in another hierarchy
            if siblingConcept.isAbstract:
                childRels = parentChildRels.fromModelObject(siblingConcept)
                checkForCalculations(val, parentChildRels, childRels, len(childRels), totalConcept, totalRel, reasonPresumedTotal, isStatementSheet, conceptsUsed, True, contributingItems)
            elif (siblingConcept in conceptsUsed and
                  siblingConcept.isNumeric and
                  siblingConcept.periodType == totalConcept.periodType):
                contributingItems.add(siblingConcept)
            if isContributingTotal:
                break
    if not nestedItems and contributingItems:
        # must check each totalFact and compatible items for a relationship set separately
        # (because different sets of sums/items could, on edge case, be in different ELRs)
        compatibleItemsFacts = defaultdict(set)
        for totalFact in val.modelXbrl.factsByQname[totalConcept.qname]:
            totalFactContext = totalFact.context
            totalFactUnit = totalFact.unit
            if (totalFactContext is not None and totalFactUnit is not None and totalFactContext.endDatetime is not None and
                (not isStatementSheet or
                 (val.requiredContext is None or
                  val.requiredContext.startDatetime <= totalFactContext.endDatetime <= val.requiredContext.endDatetime))):
                compatibleItemConcepts = set()
                compatibleFacts = {totalFact}
                for itemConcept in contributingItems:
                    for itemFact in val.modelXbrl.factsByQname[itemConcept.qname]:
                        if (totalFactContext.isEqualTo(itemFact.context) and
                            totalFactUnit.isEqualTo(itemFact.unit)):
                            compatibleItemConcepts.add(itemConcept)
                            compatibleFacts.add(itemFact)
                if len(compatibleItemConcepts) >= 2: # 6.15.2 requires 2 or more line items along with their net or total
                    compatibleItemsFacts[frozenset(compatibleItemConcepts)].update(compatibleFacts)
        for compatibleItemConcepts, compatibleFacts in compatibleItemsFacts.items():
            foundSummationItemSet = False
            leastMissingItemsSet = compatibleItemConcepts
            for ELR in val.summationItemRelsSetAllELRs.linkRoleUris:
                relSet = val.modelXbrl.relationshipSet(XbrlConst.summationItem,ELR)
                missingItems = (compatibleItemConcepts -
                                frozenset(r.toModelObject
                                          for r in relSet.fromModelObject(totalConcept)))
                # may be slow, but must remove sibling or descendants to avoid annoying false positives
                # such as in http://www.sec.gov/Archives/edgar/data/1341439/000119312512129918/orcl-20120229.xml
                missingItems -= set(concept
                                    for concept in missingItems
                                    if relSet.isRelated(totalConcept, "sibling-or-descendant", concept))
                # items not required in sum
                unrequiredItems = set(concept
                                      for concept in missingItems
                                      if concept.name in ("CommitmentsAndContingencies"))
                missingItems -= unrequiredItems
                if missingItems:
                    if len(missingItems) < len(leastMissingItemsSet):
                        leastMissingItemsSet = missingItems
                else:
                    foundSummationItemSet = True
            '''
            # testing with DH (merge all calc ELRs instead of isolating calc ELRs)...
            relSet = val.modelXbrl.relationshipSet(XbrlConst.summationItem)
            missingItems = (compatibleItemConcepts -
                            frozenset(r.toModelObject
                                      for r in relSet.fromModelObject(totalConcept)))
            foundSummationItemSet = len(missingItems) == 0
            '''
            if not foundSummationItemSet:
                linkroleDefinition = val.modelXbrl.roleTypeDefinition(contributingRel.linkrole)
                reasonIssueIsWarning = ""
                msgCode = "ERROR-SEMANTIC"
                if isStatementSheet:
                    errs = ("EFM.6.15.02,6.13.02,6.13.03", "GFM.2.06.02,2.05.02,2.05.03")
                    msg = _("Financial statement calculation relationship missing from total concept to item concepts, based on required presentation of line items and totals.  "
                            "%(reasonIssueIsWarning)s"
                            "\n\nPresentation link role: \n%(linkrole)s \n%(linkroleDefinition)s. "
                            "\n\nTotal concept: \n%(conceptSum)s.  "
                            "\n\nReason presumed total: \n%(reasonPresumedTotal)s.  "
                            "\n\nSummation items missing: \n%(missingConcepts)s.  "
                            "\n\nExpected item concepts: \n%(itemConcepts)s.  "
                            "\n\nCorresponding facts in contexts: \n%(contextIDs)s\n")
                else:
                    errs = ("EFM.6.15.03,6.13.02,6.13.03", "GFM.2.06.03,2.05.02,2.05.03")
                    msg = _("Notes calculation relationship missing from total concept to item concepts, based on required presentation of line items and totals. "
                            "%(reasonIssueIsWarning)s"
                            "\n\nPresentation link role: \n%(linkrole)s \n%(linkroleDefinition)s."
                            "\n\nTotal concept: \n%(conceptSum)s.  "
                            "\n\nReason presumed total: \n%(reasonPresumedTotal)s.  "
                            "\n\nSummation items missing \n%(missingConcepts)s.  "
                            "\n\nExpected item concepts \n%(itemConcepts)s.  "
                            "\n\nCorresponding facts in contexts: \n%(contextIDs)s\n")
                # cases causing this issue to be a warning instead of an error
                if all(f.isNil for f in compatibleFacts if f.concept in leastMissingItemsSet):
                    reasonIssueIsWarning = _("\n\nMissing items are nil, which doesn't affect validity but may impair analysis of concept semantics from calculation relationships.  ")
                    msgCode = "WARNING-SEMANTIC"
                    errs = tuple(e + '.missingItemsNil' for e in errs)
                if "parenthetical" in linkroleDefinition.lower():
                    reasonIssueIsWarning += _("\n\nLink role is parenthetical.  ")
                    msgCode = "WARNING-SEMANTIC"
                    errs = tuple(e + '.parenthetical' for e in errs)
                """@messageCatalog=[
[["EFM.6.15.02,6.13.02,6.13.03", "GFM.2.06.02,2.05.02,2.05.03"],
"Financial statement calculation relationship missing from total concept to item concepts, based on required presentation of line items and totals.
%(reasonIssueIsWarning)s

Presentation link role:
%(linkrole)s
%(linkroleDefinition)s.

Total concept:
%(conceptSum)s.

Reason presumed total: n%(reasonPresumedTotal)s.

Summation items missing n%(missingConcepts)s.

Expected item concepts
%(itemConcepts)s.

Corresponding facts in contexts:
%(contextIDs)s
"],
[["EFM.6.15.03,6.13.02,6.13.03", "GFM.2.06.03,2.05.02,2.05.03"],
"Notes calculation relationship missing from total concept to item concepts, based on required presentation of line items and totals.
%(reasonIssueIsWarning)s

Presentation link role:
%(linkrole)s
%(linkroleDefinition)s.

Total concept:
%(conceptSum)s.

Reason presumed total:
%(reasonPresumedTotal)s.

Summation items missing
%(missingConcepts)s.

Expected item concepts
%(itemConcepts)s.

Corresponding facts in contexts:
%(contextIDs)s"]]"""
                val.modelXbrl.log(msgCode, errs, msg,
                    modelObject=[totalConcept, totalRel, siblingConcept, contributingRel] + [f for f in compatibleFacts],
                    reasonIssueIsWarning=reasonIssueIsWarning,
                    conceptSum=totalConcept.qname, linkrole=contributingRel.linkrole, linkroleDefinition=linkroleDefinition,
                    reasonPresumedTotal=reasonPresumedTotal,
                    itemConcepts=', \n'.join(sorted(set(str(c.qname) for c in compatibleItemConcepts))),
                    missingConcepts = ', \n'.join(sorted(set(str(c.qname) for c in leastMissingItemsSet))),
                    contextIDs=', '.join(sorted(set(f.contextID for f in compatibleFacts))))
            leastMissingItemsSet = None #dereference, can't delete with Python 3.1
            del foundSummationItemSet
        del compatibleItemsFacts # dereference object references



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import ModelDocument, ViewFile
from arelle.ViewUtil import sortCountExpected
import os

def viewTests(modelXbrl, outfile, cols=None):
    modelXbrl.modelManager.showStatus(_("viewing Tests"))
    view = ViewTests(modelXbrl, outfile, cols)
    view.viewTestcaseIndexElement(modelXbrl.modelDocument)
    view.close()

COL_WIDTHS = {
    "Index": 12,
    "Index.1": 12,
    "Index.2": 12,
    "Index.3": 12,
    "Index.4": 12,
    "Testcase": 20,
    "Id": 10,
    "Name": 50,
    "Reference": 20,
    "ReadMeFirst": 20,
    "Status": 8,
    "Expected": 20,
    "Actual": 100}

class ViewTests(ViewFile.View):
    def __init__(self, modelXbrl, outfile, cols):
        super(ViewTests, self).__init__(modelXbrl, outfile, "Tests")
        self.cols = cols

    def viewTestcaseIndexElement(self, modelDocument, parentDocument=None, nestedDepth=0):
        if parentDocument is None: # not a nested testacases index
            self.nestedIndexDepth = 0
            self.xlsxDocNames = []
            self.xlsxTestcase = ""
            if self.cols:
                if isinstance(self.cols,str): self.cols = self.cols.replace(',',' ').split()
                unrecognizedCols = []
                for col in self.cols:
                    if col not in COL_WIDTHS:
                        unrecognizedCols.append(col)
                if unrecognizedCols:
                    self.modelXbrl.error("arelle:unrecognizedTestReportColumn",
                                         _("Unrecognized columns: %(cols)s"),
                                         modelXbrl=self.modelXbrl, cols=','.join(unrecognizedCols))
                if "Period" in self.cols:
                    i = self.cols.index("Period")
                    self.cols[i:i+1] = ["Start", "End/Instant"]
            else:
                self.cols = ["Index"]
                def determineNestedIndexDepth(doc, nestedDepth):
                    if nestedDepth > self.nestedIndexDepth:
                        self.nestedIndexDepth = nestedDepth
                        self.cols.append("Index.{}".format(nestedDepth))
                        self.xlsxDocNames.append("")
                    for referencedDocument in doc.referencesDocument.keys():
                        if referencedDocument.type == ModelDocument.Type.TESTCASESINDEX:
                            determineNestedIndexDepth(referencedDocument, nestedDepth + 1)
                determineNestedIndexDepth(modelDocument, 0)
                self.cols += ["Testcase", "Id"]
                if self.type != ViewFile.XML:
                    self.cols.append("Name")
                    self.xlsxDocNames.append("")
                self.cols += ["ReadMeFirst", "Status", "Expected", "Actual"]

            self.setColWidths([COL_WIDTHS.get(col, 8) for col in self.cols])
            self.addRow(self.cols, asHeader=True)

        if modelDocument.type in (ModelDocument.Type.TESTCASESINDEX, ModelDocument.Type.REGISTRY):
            cols = []
            attr = {}
            if self.type in (ViewFile.XLSX, ViewFile.CSV):
                self.xlsxDocName = None
            indexColName = "Index.{}".format(nestedDepth) if nestedDepth else "Index"
            for col in self.cols:
                if col == indexColName:
                    docName = os.path.basename(modelDocument.uri)
                    if parentDocument and os.path.basename(parentDocument.uri) == docName:
                        docName = os.path.basename(os.path.dirname(modelDocument.uri))
                    if self.type in (ViewFile.XLSX, ViewFile.CSV):
                        self.xlsxDocNames[nestedDepth] = docName
                    else:
                        attr["name"] = docName
                    break
                else:
                    cols.append("")
            if self.type not in (ViewFile.XLSX, ViewFile.CSV):
                self.addRow(cols, treeIndent=0, xmlRowElementName="testcaseIndex", xmlRowEltAttr=attr, xmlCol0skipElt=True)
            # sort test cases by uri
            testcases = []
            for referencedDocument, _ref in sorted(modelDocument.referencesDocument.items(),
                                             key=lambda i:i[1].referringModelObject.objectIndex if i[1] else 0):
                if referencedDocument.type in (ModelDocument.Type.TESTCASESINDEX, ModelDocument.Type.REGISTRY):
                    self.viewTestcaseIndexElement(referencedDocument, modelDocument, nestedDepth+1)
                else:
                    testcases.append((referencedDocument.uri, referencedDocument.objectId()))
            testcases.sort()
            for testcaseTuple in testcases:
                self.viewTestcase(self.modelXbrl.modelObject(testcaseTuple[1]), 1)
        elif modelDocument.type in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
            self.viewTestcase(modelDocument, 0)
        else:
            pass

    def viewTestcase(self, modelDocument, indent):
        cols = []
        attr = {}
        docName = os.path.basename(modelDocument.uri)
        if docName == "index.xml":
            docName = os.path.basename(os.path.dirname(modelDocument.uri))
        if self.type in (ViewFile.XLSX, ViewFile.CSV):
            self.xlsxTestcase = os.path.basename(docName)
        for col in self.cols:
            if col == "Testcase":
                if self.type != ViewFile.XML:
                    cols.append(docName)
                else:
                    attr["name"] = os.path.basename(modelDocument.uri)
                break
            else:
                cols.append("")
        if self.type not in (ViewFile.XLSX, ViewFile.CSV):
            self.addRow(cols, treeIndent=indent, xmlRowElementName="testcase", xmlRowEltAttr=attr, xmlCol0skipElt=True)
        for modelTestcaseVariation in getattr(modelDocument, "testcaseVariations", ()):
            self.viewTestcaseVariation(modelTestcaseVariation, indent+1)

    def viewTestcaseVariation(self, modelTestcaseVariation, indent):
        id = modelTestcaseVariation.id
        if id is None:
            id = ""
        cols = []
        attr = {}
        if self.type in (ViewFile.XLSX, ViewFile.CSV):
            cols.extend(self.xlsxDocNames)
            cols.append(self.xlsxTestcase)
            indent = 0 # excel & CSV show all columns to allow filtering
        for col in self.cols:
            if self.type in (ViewFile.XLSX, ViewFile.CSV) and (col.startswith("Index") or col == "Testcase"):
                pass # these columns added above
            elif col == "Id":
                cols.append(id or modelTestcaseVariation.name)
            elif col == "Name":
                if self.type != ViewFile.XML:
                    cols.append(modelTestcaseVariation.description or modelTestcaseVariation.name)
                else:
                    attr["name"] = modelTestcaseVariation.description or modelTestcaseVariation.name
            elif col == "Reference":
                cols.append(modelTestcaseVariation.reference)
            elif col == "ReadMeFirst":
                cols.append(" ".join(str(uri) for uri in modelTestcaseVariation.readMeFirstUris))
            elif col == "Status":
                cols.append(modelTestcaseVariation.status)
            elif col == "Expected":
                _exp = sortCountExpected(modelTestcaseVariation.expected)
                cols.append(", ".join(str(e) for e in _exp) if isinstance(_exp, list) else _exp)
            elif col == "Actual":
                cols.append(", ".join(str(code) for code in modelTestcaseVariation.actual))
            else:
                cols.append("")
        self.addRow(cols, treeIndent=indent, xmlRowElementName="variation", xmlRowEltAttr=attr, xmlCol0skipElt=False)



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
import os, datetime
import regex as re
from tkinter import Menu, constants, BooleanVar
from arelle import ViewWinTree, ModelDtsObject, ModelInstanceObject, XbrlConst, XmlUtil
from arelle.XbrlConst import conceptNameLabelRole

stripXmlPattern = re.compile(r"<.*?>", re.DOTALL)
decEntityPattern = re.compile(r"&#([0-9]+);")
hexEntityPattern = re.compile(r"&#x([0-9]+);")
stripWhitespacePattern = re.compile(r"\s\s+")

def viewFacts(modelXbrl, tabWin, header="Fact Table", arcrole=XbrlConst.parentChild, linkrole=None, linkqname=None, arcqname=None, lang=None, expandAll=False):
    modelXbrl.modelManager.showStatus(_("viewing relationships {0}").format(os.path.basename(arcrole)))
    view = ViewFactTable(modelXbrl, tabWin, header, arcrole, linkrole, linkqname, arcqname, lang, expandAll)
    view.ignoreDims = BooleanVar(value=False)
    view.showDimDefaults = BooleanVar(value=False)
    view.view()
    view.treeView.bind("<<TreeviewSelect>>", view.treeviewSelect, '+')
    view.treeView.bind("<ButtonRelease-1>", view.treeviewClick, '+')
    view.treeView.bind("<Enter>", view.treeviewEnter, '+')
    view.treeView.bind("<Leave>", view.treeviewLeave, '+')

    # languages menu
    menu = view.contextMenu()
    optionsMenu = Menu(view.viewFrame, tearoff=0)
    view.ignoreDims.trace("w", view.viewReloadDueToMenuAction)
    optionsMenu.add_checkbutton(label=_("Ignore Dimensions"), underline=0, variable=view.ignoreDims, onvalue=True, offvalue=False)
    menu.add_cascade(label=_("Options"), menu=optionsMenu, underline=0)
    view.menuAddExpandCollapse()
    view.menuAddClipboard()
    view.menuAddLangs()
    view.menuAddLabelRoles(includeConceptName=True)
    #saveMenu = Menu(view.viewFrame, tearoff=0)
    #saveMenu.add_command(label=_("HTML file"), underline=0, command=lambda: view.modelXbrl.modelManager.cntlr.fileSave(view=view, fileType="html"))
    #menu.add_cascade(label=_("Save"), menu=saveMenu, underline=0)

class ViewFactTable(ViewWinTree.ViewTree):
    def __init__(self, modelXbrl, tabWin, header, arcrole, linkrole=None, linkqname=None, arcqname=None, lang=None, expandAll=False):
        super(ViewFactTable, self).__init__(modelXbrl, tabWin, header, True, lang)
        self.arcrole = arcrole
        self.linkrole = linkrole
        self.linkqname = linkqname
        self.arcqname = arcqname
        self.expandAllOnFirstDisplay = expandAll

    def viewReloadDueToMenuAction(self, *args):
        self.view()

    def view(self):
        self.blockSelectEvent = 1
        self.blockViewModelObject = 0
        self.tag_has = defaultdict(list) # temporary until Tk 8.6
        # relationship set based on linkrole parameter, to determine applicable linkroles
        relationshipSet = self.modelXbrl.relationshipSet(self.arcrole, self.linkrole, self.linkqname, self.arcqname)
        if not relationshipSet:
            self.modelXbrl.modelManager.addToLog(_("no relationships for {0}").format(self.arcrole))
            return False
        # consider facts in the relationshipSet (only)
        contexts = set()
        self.conceptFacts = defaultdict(list)
        if self.linkrole and self.modelXbrl.roleTypes[self.linkrole] and hasattr(self.modelXbrl.roleTypes[self.linkrole][0], "_tableFacts"):
            for fact in self.modelXbrl.roleTypes[self.linkrole][0]._tableFacts:
                self.conceptFacts[fact.qname].append(fact)
                if fact.context is not None:
                    contexts.add(fact.context)
        else:
            for fact in self.modelXbrl.facts:
                if relationshipSet.fromModelObject(fact.concept) or relationshipSet.toModelObject(fact.concept):
                    self.conceptFacts[fact.qname].append(fact)
                    if fact.context is not None:
                        contexts.add(fact.context)
        # sort contexts by period
        self.periodContexts = defaultdict(set)
        contextStartDatetimes = {}
        ignoreDims = self.ignoreDims.get()
        showDimDefaults = self.showDimDefaults.get()
        for context in contexts:
            if context is None or context.endDatetime is None:
                contextkey = "missing period"
            elif ignoreDims:
                if context.isForeverPeriod:
                    contextkey = datetime.datetime(datetime.MINYEAR,1,1)
                else:
                    contextkey = context.endDatetime
            else:
                if context.isForeverPeriod:
                    contextkey = "forever"
                else:
                    contextkey = (context.endDatetime - datetime.timedelta(days=1)).strftime("%Y-%m-%d")

                values = []
                dims = context.qnameDims
                if len(dims) > 0:
                    for dimQname in sorted(dims.keys(), key=lambda d: str(d)):
                        dimvalue = dims[dimQname]
                        if dimvalue.isExplicit:
                            values.append(dimvalue.member.label(self.labelrole,lang=self.lang)
                                          if dimvalue.member is not None
                                          else str(dimvalue.memberQname))
                        else:
                            values.append(XmlUtil.innerText(dimvalue.typedMember))

                nonDimensions = context.nonDimValues("segment") + context.nonDimValues("scenario")
                if len(nonDimensions) > 0:
                    for element in sorted(nonDimensions, key=lambda e: e.localName):
                        values.append(XmlUtil.innerText(element))

                if len(values) > 0:

                    contextkey += " - " + ', '.join(values)

            objectId = context.objectId()
            self.periodContexts[contextkey].add(objectId)
            if context.isStartEndPeriod:
                contextStartDatetimes[objectId] = context.startDatetime
        self.periodKeys = list(self.periodContexts.keys())
        self.periodKeys.sort()
        # set up treeView widget and tabbed pane
        self.treeView.column("#0", width=300, anchor="w")
        self.treeView.heading("#0", text="Concept")
        columnIds = []
        columnIdHeadings = []
        self.contextColId = {}
        self.startdatetimeColId = {}
        self.numCols = 1
        for periodKey in self.periodKeys:
            colId = "#{0}".format(self.numCols)
            columnIds.append(colId)
            columnIdHeadings.append((colId,periodKey))
            for contextId in self.periodContexts[periodKey]:
                self.contextColId[contextId] = colId
                if contextId in contextStartDatetimes:
                    self.startdatetimeColId[contextStartDatetimes[contextId]] = colId
            self.numCols += 1
        self.treeView["columns"] = columnIds
        for colId, colHeading in columnIdHeadings:
            self.treeView.column(colId, width=100, anchor="w")
            if ignoreDims:
                if colHeading.year == datetime.MINYEAR:
                    date = "forever"
                else:
                    date = (colHeading - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
                self.treeView.heading(colId, text=date)
            else:
                self.treeView.heading(colId, text=colHeading)

        # fact rendering
        self.clearTreeView()
        self.rowColFactId = {}
        # root node for tree view
        self.id = 1
        # sort URIs by definition
        linkroleUris = []
        relationshipSet = self.modelXbrl.relationshipSet(self.arcrole, self.linkrole, self.linkqname, self.arcqname)
        if self.linkrole:
            roleType = self.modelXbrl.roleTypes[self.linkrole][0]
            linkroleUris.append(((roleType.genLabel(lang=self.lang, strip=True) or
                                  roleType.definition or
                                  linkroleUri), self.linkrole, roleType.objectId(self.id)))
            self.id += 1
        else:
            for linkroleUri in relationshipSet.linkRoleUris:
                modelRoleTypes = self.modelXbrl.roleTypes.get(linkroleUri)
                if modelRoleTypes:
                    roledefinition = (modelRoleTypes[0].genLabel(lang=self.lang, strip=True) or modelRoleTypes[0].definition or linkroleUri)
                    roleId = modelRoleTypes[0].objectId(self.id)
                else:
                    roledefinition = linkroleUri
                    roleId = "node{0}".format(self.id)
                self.id += 1
                linkroleUris.append((roledefinition, linkroleUri, roleId))
            linkroleUris.sort()
        # for each URI in definition order
        for linkroleUriTuple in linkroleUris:
            linknode = self.treeView.insert("", "end", linkroleUriTuple[2], text=linkroleUriTuple[0], tags=("ELR",))
            linkRelationshipSet = self.modelXbrl.relationshipSet(self.arcrole, linkroleUriTuple[1], self.linkqname, self.arcqname)
            for rootConcept in linkRelationshipSet.rootConcepts:
                node = self.viewConcept(rootConcept, rootConcept, "", self.labelrole, linknode, 1, linkRelationshipSet, set())

        if self.expandAllOnFirstDisplay:
            self.expandAll()

        return True

    def viewConcept(self, concept, modelObject, labelPrefix, preferredLabel, parentnode, n, relationshipSet, visited):
        # bad relationship could identify non-concept or be None
        if (not isinstance(concept, ModelDtsObject.ModelConcept) or
            concept.substitutionGroupQname == XbrlConst.qnXbrldtDimensionItem):
            return
        childnode = self.treeView.insert(parentnode, "end", modelObject.objectId(self.id),
                    text=labelPrefix + concept.label(preferredLabel,lang=self.lang,linkroleHint=relationshipSet.linkrole),
                    tags=("odd" if n & 1 else "even",))
        self.setRowFacts(childnode,concept,preferredLabel)
        self.id += 1
        self.tag_has[modelObject.objectId()].append(childnode)
        if isinstance(modelObject, ModelDtsObject.ModelRelationship):
            self.tag_has[modelObject.toModelObject.objectId()].append(childnode)
        if concept not in visited:
            visited.add(concept)
            for i, modelRel in enumerate(relationshipSet.fromModelObject(concept)):
                nestedRelationshipSet = relationshipSet
                targetRole = modelRel.targetRole
                if self.arcrole == XbrlConst.summationItem:
                    childPrefix = "({:0g}) ".format(modelRel.weight) # format without .0 on integer weights
                elif targetRole is None or len(targetRole) == 0:
                    targetRole = relationshipSet.linkrole
                    childPrefix = ""
                else:
                    nestedRelationshipSet = self.modelXbrl.relationshipSet(self.arcrole, targetRole, self.linkqname, self.arcqname)
                    childPrefix = "(via targetRole) "
                toConcept = modelRel.toModelObject
                if toConcept in visited:
                    childPrefix += "(loop)"
                labelrole = modelRel.preferredLabel
                if not labelrole or self.labelrole == conceptNameLabelRole:
                    labelrole = self.labelrole
                self.viewConcept(toConcept, modelRel, childPrefix, labelrole, childnode, n + i + 1, nestedRelationshipSet, visited)
            visited.remove(concept)

    def setRowFacts(self, node, concept, preferredLabel):
        for fact in self.conceptFacts[concept.qname]:
            try:
                colId = self.contextColId[fact.context.objectId()]
                # special case of start date, pick column corresponding
                if preferredLabel == XbrlConst.periodStartLabel:
                    date = fact.context.instantDatetime
                    if date:
                        if date in self.startdatetimeColId:
                            colId = self.startdatetimeColId[date]
                        else:
                            continue # not shown on this row (belongs on end period label row
                self.treeView.set(node,
                                  colId,
                                  fact.effectiveValue if concept is None
                                  or concept.isNumeric else
                                  hexEntityPattern.sub(
                                    lambda m: chr(int('0x'+m.group(1),16)),
                                    decEntityPattern.sub(
                                      lambda m: chr(int(m.group(1),10)),
                                      stripWhitespacePattern.sub(" ",
                                        stripXmlPattern.sub(" ", fact.stringValue)))
                                    ).strip(),
                                )
                factObjectId = fact.objectId()
                self.tag_has[factObjectId].append(node)
                self.rowColFactId[node + colId] = factObjectId
            except AttributeError:  # not a fact or no concept
                pass

    def treeviewEnter(self, *args):
        self.blockSelectEvent = 0

    def treeviewLeave(self, *args):
        self.blockSelectEvent = 1

    def treeviewClick(self, event):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            colId = self.treeView.identify_column(event.x)
            modelObjectId = self.treeView.identify_row(event.y)
            if colId != "" and modelObjectId != "":
                if colId == "#0":
                    self.modelXbrl.viewModelObject(modelObjectId)
                else:
                    factId = self.rowColFactId.get(modelObjectId + colId)
                    self.modelXbrl.viewModelObject(factId)
            self.blockViewModelObject -= 1

    def treeviewSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            #self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                if isinstance(modelObject, ModelDtsObject.ModelRoleType):
                    self.linkrole = modelObject.roleURI
                    self.view()
                    modelObject = None
                # get concept of fact or toConcept of relationship, role obj if roleType
                if not isinstance(modelObject, ModelInstanceObject.ModelFact):
                    modelObject = modelObject.viewConcept
                if modelObject is not None:
                    items = self.tag_has.get(modelObject.objectId())
                    if items is not None and self.treeView.exists(items[0]):
                        self.treeView.see(items[0])
                        self.treeView.selection_set(items[0])
            except (AttributeError, KeyError):
                self.treeView.selection_set(())
            if self.blockViewModelObject > 0:
                self.blockViewModelObject -= 1
